from collections import defaultdict
from textwrap import dedent

from drf_spectacular.utils import extend_schema_serializer
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.documents.models import Corpus, Entity, EntityLink, EntityRole, EntityType, TranscriptionEntity
from arkindex.documents.serializers.light import CorpusLightSerializer, EntityTypeLightSerializer
from arkindex.documents.serializers.ml import WorkerRunSummarySerializer
from arkindex.process.models import WorkerRun
from arkindex.project.serializer_fields import ForbiddenField
from arkindex.project.validators import ForbiddenValidator, XorValidator


class EntityTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = EntityType
        fields = (
            'id',
            'name',
            'color',
            'corpus'
        )
        read_only_fields = (
            'id',
            'corpus'
        )

    def to_internal_value(self, data):
        # Get an error if a request tries to change an entity type's corpus
        corpus = data.get('corpus')
        if self.instance and corpus:
            raise ValidationError({'corpus': ['It is not possible to update an Entity Type\'s corpus.']})
        data = super().to_internal_value(data)
        return data

    def validate(self, data):
        name = data.get('name') if 'name' in data else self.instance.name
        corpus = data.get('corpus') if 'corpus' in data else self.instance.corpus
        existing_type = corpus.entity_types.filter(name=name).using('default').first()
        if existing_type is not None and (not self.instance or existing_type.id != self.instance.id):
            raise ValidationError({'name': [f'Entity Type with name {name} already exists in corpus {corpus}.']})
        return data


class EntityTypeCreateSerializer(EntityTypeSerializer):
    class Meta:
        model = EntityType
        fields = EntityTypeSerializer.Meta.fields
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields['corpus'].queryset = Corpus.objects.admin(self.context['request'].user)


@extend_schema_serializer(deprecate_fields=('worker_version_id', ))
class BaseEntitySerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EntityTypeLightSerializer()
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    worker_version_id = serializers.PrimaryKeyRelatedField(read_only=True)
    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'worker_version_id',
            'worker_run',
        )
        read_only_fields = (
            'id',
            'worker_version_id',
            'worker_run',
        )


class EntityRoleSerializer(serializers.ModelSerializer):
    """
    Serialize a role between two types of entity
    """
    parent_type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={'base_template': 'input.html'},
        source='parent_type',
    )
    child_type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={'base_template': 'input.html'},
        source='child_type',
    )

    class Meta:
        model = EntityRole
        fields = (
            'id',
            'parent_name',
            'child_name',
            'parent_type_id',
            'child_type_id'
        )

    def validate(self, data):
        errors = defaultdict(list)
        assert 'corpus' not in data
        assert self.context.get('request') is not None
        corpus_id = self.context['request'].parser_context['kwargs']['pk']
        corpus = Corpus.objects.writable(self.context['request'].user).filter(id=corpus_id).first()
        if corpus is None:
            raise serializers.ValidationError({
                'corpus': ['You do not have write access to this corpus'],
                'id': corpus_id,
            })
        parent_type = data.get('parent_type')
        child_type = data.get('child_type')
        if parent_type.corpus_id != corpus.id:
            errors['parent_type_id'].append(f'Type {parent_type.id} does not exist in corpus {corpus}.')
        if child_type.corpus_id != corpus.id:
            errors['child_type_id'].append(f'Type {child_type.id} does not exist in corpus {corpus}.')

        if errors:
            raise ValidationError(errors)

        data['corpus'] = corpus
        return data


class CreateEntityRoleErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="The corpus ID.")
    corpus = serializers.ListField(child=serializers.CharField(), required=False, help_text="Errors that occurred during corpus ID field validation.")


class EntityLinkSerializer(serializers.ModelSerializer):
    """
    Serialize an entity link with its child, parent and role
    """
    parent = BaseEntitySerializer()
    child = BaseEntitySerializer()
    role = EntityRoleSerializer()

    class Meta:
        model = EntityLink
        fields = (
            'id',
            'parent',
            'child',
            'role'
        )


@extend_schema_serializer(
    deprecate_fields=('type_name', 'worker_version_id')
)
class EntitySerializer(BaseEntitySerializer):
    """
    Serialize an entity with its metadata
    """
    corpus = CorpusLightSerializer(read_only=True)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)
    # When updating an entity, the type can be set either by using its EntityType UUID, or its name
    # (in which case the serializer checks that an EntityType with this name exists in the corpus)
    type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={'base_template': 'input.html'},
        required=False,
        write_only=True
    )
    type_name = serializers.CharField(
        required=False,
        write_only=True,
        help_text=('The name of an existing entity type in the entity\'s corpus. '
                   'This field is deprecated: use `type_id` instead.')
    )
    type = EntityTypeLightSerializer(read_only=True)

    class Meta:
        model = Entity
        fields = BaseEntitySerializer.Meta.fields + (
            'corpus',
            'children',
            'parents',
            'type_id',
            'type_name'
        )
        read_only_fields = BaseEntitySerializer.Meta.read_only_fields = (
            'corpus',
            'children',
            'parents',
        )
        validators = [
            XorValidator('type_name', 'type_id', required=True)
        ]

    def validate(self, data):
        errors = defaultdict(list)

        if data.get('validated'):
            data['moderator'] = self.context['request'].user
        corpus = self.instance.corpus
        type = data.get('type_id')
        type_name = data.get('type_name')
        if type_name:
            try:
                data['type'] = EntityType.objects.get(corpus=corpus, name=type_name)
            except EntityType.DoesNotExist:
                errors['type_name'].append(f'EntityType {type_name} does not exist in corpus {corpus.name}.')
        elif type:
            if type.corpus_id != corpus.id:
                errors['type_id'].append(f'EntityType {type.id} does not exist in corpus {corpus.name}.')
            data['type'] = type

        if errors:
            raise ValidationError(errors)
        return data


@extend_schema_serializer(
    deprecate_fields=('type_name', 'worker_version', 'worker_version_id')
)
class EntityCreateSerializer(BaseEntitySerializer):
    """
    Serialize an entity with possible parents and children
    """
    corpus = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        style={'base_template': 'input.html'},
    )
    metas = serializers.HStoreField(child=serializers.CharField(), required=False)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)
    worker_version = serializers.UUIDField(
        allow_null=True,
        required=False,
        source='worker_version_id',
        validators=[
            ForbiddenValidator(),
        ],
        help_text=dedent("""
            ID of a WorkerVersion that created this entity.

            Creating new entities with a WorkerVersion is forbidden. Use `worker_run_id` instead.
        """),
    )
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        default=None,
        source='worker_run',
        style={'base_template': 'input.html'},
    )
    # When creating an entity, the type can be set either by using its EntityType UUID, or its name
    # (in which case the serializer checks that an EntityType with this name exists in the corpus)
    type_id = serializers.PrimaryKeyRelatedField(
        queryset=EntityType.objects.all(),
        style={'base_template': 'input.html'},
        required=False,
    )
    type_name = serializers.CharField(
        required=False,
        help_text=('The name of an existing entity type in the target corpus. '
                   'This field is deprecated: use `type_id` instead.')
    )

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type_name',
            'type_id',
            'metas',
            'validated',
            'corpus',
            'parents',
            'children',
            'worker_version',
            'worker_run_id'
        )
        read_only_fields = (
            'id',
            'children',
            'parents',
        )
        validators = [
            XorValidator('type_name', 'type_id', required=True)
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora

    def validate(self, data):
        errors = defaultdict(list)

        worker_run = data.get('worker_run')
        if worker_run:
            data['worker_version_id'] = worker_run.version_id

        corpus = data.get('corpus')
        type = data.get('type_id')
        type_name = data.get('type_name')
        if type_name:
            try:
                data['type'] = EntityType.objects.get(corpus=corpus, name=type_name)
            except EntityType.DoesNotExist:
                errors['type_name'].append(f'EntityType {type_name} does not exist in corpus {corpus.name}.')
        else:
            if type.corpus_id != corpus.id:
                errors['type_id'].append(f'EntityType {type.id} does not exist in corpus {corpus.name}.')
            data['type'] = type

        if errors:
            raise ValidationError(errors)

        return data


class EntityLinkCreateSerializer(EntityLinkSerializer):
    """
    Serialize an entity with a possible parents and children
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.none(),
        style={'base_template': 'input.html'},
    )
    child = serializers.PrimaryKeyRelatedField(
        queryset=Entity.objects.none(),
        style={'base_template': 'input.html'},
    )
    role = serializers.PrimaryKeyRelatedField(
        queryset=EntityRole.objects.none(),
        style={'base_template': 'input.html'},
    )

    class Meta:
        model = EntityLink
        fields = EntityLinkSerializer.Meta.fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        entities = Entity.objects.all().filter(corpus__in=corpora)
        roles = EntityRole.objects.all().filter(corpus__in=corpora)
        self.fields['parent'].queryset = entities
        self.fields['child'].queryset = entities
        self.fields['role'].queryset = roles

    def validate(self, data):
        data = super().validate(data)
        link = EntityLink(**data)
        link.full_clean()
        return data


class TranscriptionEntityCreateSerializer(serializers.ModelSerializer):
    """
    Serialise the link between an entity and a transcription
    Transcription element must be passed in serializer context
    """
    offset = serializers.IntegerField(min_value=0)
    length = serializers.IntegerField(min_value=1)
    worker_version_id = ForbiddenField()
    worker_run_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerRun.objects.all(),
        write_only=True,
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
        source='worker_run'
    )
    confidence = serializers.FloatField(min_value=0, max_value=1, default=None)

    worker_run = WorkerRunSummarySerializer(read_only=True, allow_null=True)

    class Meta:
        model = TranscriptionEntity
        fields = (
            'entity',
            'offset',
            'length',
            'worker_version_id',
            'worker_run_id',
            'worker_run',
            'confidence',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        self.fields['entity'].queryset = Entity.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user)
        )

    def validate(self, data):
        data = super().validate(data)
        errors = {}

        data['transcription'] = self.context['transcription']
        if not data['transcription']:
            errors['transcription'] = ['invalid UUID or Corpus write-access is forbidden']

        worker_run = data.get('worker_run')
        if worker_run is not None:
            data['worker_version_id'] = worker_run.version_id

        existing_transcription_entities = TranscriptionEntity.objects.filter(
            transcription=data['transcription'],
            entity=data['entity'],
            offset=data['offset'],
            length=data['length'],
            worker_run=worker_run,
        )
        if existing_transcription_entities.exists():
            if worker_run:
                errors['__all__'] = ['This entity is already linked to this transcription by this worker run at this position.']
            else:
                errors['__all__'] = ['This entity is already linked to this transcription at this position.']

        if errors:
            raise serializers.ValidationError(errors)

        return data


@extend_schema_serializer(deprecate_fields=('worker_version_id', ))
class TranscriptionEntitySerializer(TranscriptionEntityCreateSerializer):
    """
    Serialize the link between a transcription and a detailed entity
    """
    worker_version_id = serializers.UUIDField(read_only=True)
    entity = BaseEntitySerializer()
