import os
from pathlib import Path
from unittest.mock import patch

from django.conf import settings
from django.core.checks import Critical, Error, Warning
from django.test import TestCase, override_settings
from django.urls import path


@override_settings(ALL_CHECKS=True)
class ChecksTestCase(TestCase):

    def test_api_urls_check(self):
        '''
        Check should always be OK
        '''
        from arkindex.project.api_v1 import api
        from arkindex.project.checks import api_urls_check

        self.assertListEqual(api_urls_check(), [])

        # Add an invalid url
        bad_url = path('bad', lambda: None, name='bad-url')
        api.append(bad_url)

        # Should trigger the check
        self.assertListEqual(
            api_urls_check(),
            [
                Error(
                    'This API URL misses a mandatory ending /',
                    hint=str(bad_url),
                    id='arkindex.E001',
                )
            ]
        )

    def test_local_imageserver_check(self):
        """
        Test the local imageserver existence check
        """
        from arkindex.images.models import ImageServer
        from arkindex.project.checks import local_imageserver_check
        self.assertEqual(ImageServer.objects.count(), 0)

        with self.settings(LOCAL_IMAGESERVER_ID=42):
            self.assertListEqual(local_imageserver_check(), [
                Warning(
                    'Local ImageServer with ID 42 does not exist',
                    hint='settings.LOCAL_IMAGESERVER_ID = 42',
                    id='arkindex.W009',
                ),
            ])

        srv = ImageServer.objects.create(display_name="Test Local Server", url="http://nowhere")

        with self.settings(LOCAL_IMAGESERVER_ID=srv.id):
            self.assertListEqual(local_imageserver_check(), [])

        srv.delete()

    @override_settings()
    @patch('arkindex.project.checks.parse_recipe')
    def test_ponos_recipe_check(self, parse_mock):
        from arkindex.project.checks import ponos_recipe_check
        self.assertListEqual(ponos_recipe_check(), [])

        parse_mock.side_effect = AssertionError('oh no')
        self.assertListEqual(ponos_recipe_check(), [Error(
            'Base Ponos recipe validation failed: oh no',
            hint='settings.PONOS_RECIPE',
            id='arkindex.E007',
        )])

        settings.CONFIG_PATH = Path('/somewhere/config.yml')
        del settings.PONOS_RECIPE['env']
        parse_mock.side_effect = None
        self.assertListEqual(ponos_recipe_check(), [
            Warning(
                'The ARKINDEX_API_URL environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
            Warning(
                'The ARKINDEX_API_TOKEN environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
            Warning(
                'The ARKINDEX_API_CSRF_COOKIE environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
        ])

    @override_settings()
    def test_gitlab_oauth_check(self):
        from arkindex.project.checks import gitlab_oauth_check

        del settings.GITLAB_APP_ID
        del settings.GITLAB_APP_SECRET

        self.assertListEqual(gitlab_oauth_check(), [
            Warning(
                'GitLab app ID is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_ID = None',
                id='arkindex.W003',
            ),
            Warning(
                'GitLab app secret is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_SECRET = None',
                id='arkindex.W004',
            ),
        ])

        settings.GITLAB_APP_ID = '1234'
        settings.GITLAB_APP_SECRET = 's3kr3t'

        self.assertListEqual(gitlab_oauth_check(), [])

    @override_settings()
    def test_s3_check(self):
        from arkindex.project.checks import s3_check

        del settings.AWS_ACCESS_KEY
        del settings.AWS_SECRET_KEY
        del settings.AWS_THUMBNAIL_BUCKET
        del settings.AWS_STAGING_BUCKET
        del settings.AWS_EXPORT_BUCKET
        self.assertCountEqual(s3_check(), [
            Error(
                'AWS access key ID is missing; all S3-related features will fail.',
                hint='settings.AWS_ACCESS_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'AWS secret key is missing; all S3-related features will fail.',
                hint='settings.AWS_SECRET_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 thumbnails bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_THUMBNAIL_BUCKET = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 staging bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_STAGING_BUCKET = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 export bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_EXPORT_BUCKET = None',
                id='arkindex.E011',
            ),
        ])

        settings.AWS_ACCESS_KEY = 'key'
        settings.AWS_SECRET_KEY = 's3kr3t'
        settings.AWS_THUMBNAIL_BUCKET = 'Thumbs.db'
        settings.AWS_STAGING_BUCKET = 'buckette'
        settings.AWS_EXPORT_BUCKET = 'devnull'
        self.assertListEqual(s3_check(), [])

    @override_settings()
    def test_public_hostname_check(self):
        from arkindex.project.checks import public_hostname_check

        settings.PUBLIC_HOSTNAME = None
        self.assertListEqual(public_hostname_check(), [
            Warning(
                'The public_hostname setting is missing; absolute URLs may not be correctly generated.',
                hint='settings.PUBLIC_HOSTNAME',
                id='arkindex.W008',
            )
        ])

        settings.PUBLIC_HOSTNAME = 'https://darkindex.lol'
        self.assertListEqual(public_hostname_check(), [])

    @patch('arkindex.project.checks.sys.version_info')
    def test_python_version_check(self, version_info_mock):
        from arkindex.project.checks import python_version_check

        version_info_mock.minor = 7
        self.assertListEqual(python_version_check(), [
            Critical(
                'Arkindex requires Python 3.8 or later.',
                id='arkindex.C001',
            )
        ])

        version_info_mock.minor = 8
        self.assertListEqual(python_version_check(), [])

        version_info_mock.minor = 9
        self.assertListEqual(python_version_check(), [])

    @override_settings()
    def test_ingest_check(self):
        from arkindex.images.models import ImageServer
        from arkindex.project.checks import ingest_check

        settings.INGEST_S3_ACCESS_KEY = None
        settings.INGEST_S3_SECRET_KEY = None
        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = None
        settings.INGEST_IMAGESERVER_ID = None
        self.assertCountEqual(ingest_check(), [
            Warning(
                'INGEST_S3_ACCESS_KEY is not set. The S3 ingest feature will not work.',
                id='arkindex.W010',
            ),
            Warning(
                'INGEST_S3_SECRET_KEY is not set. The S3 ingest feature will not work.',
                id='arkindex.W010',
            ),
            Warning(
                'Neither INGEST_S3_REGION nor INGEST_S3_ENDPOINT are set. The S3 ingest feature will not work.',
                id='arkindex.W010',
            ),
            Warning(
                'INGEST_IMAGESERVER_ID is not set. The S3 ingest feature will not work.',
                id='arkindex.W010',
            ),
        ])

        settings.INGEST_S3_ACCESS_KEY = 'key'
        settings.INGEST_S3_SECRET_KEY = 's3kr3t'
        settings.INGEST_S3_ENDPOINT = 'something'
        settings.INGEST_IMAGESERVER_ID = 42
        self.assertListEqual(ingest_check(), [
            Warning(
                'ImageServer with ID 42 does not exist. The S3 ingest feature will not work.',
                hint='settings.INGEST_IMAGESERVER_ID = 42',
                id='arkindex.W010',
            )
        ])

        ImageServer.objects.create(id=42, display_name='Test Server', url='https://null.teklia.com/dev/null')
        self.assertListEqual(ingest_check(), [])

        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = 'somewhere'
        self.assertListEqual(ingest_check(), [])

    @override_settings()
    @patch('os.path.exists', return_value=True)
    @patch.dict(os.environ, {
        'AWS_ACCESS_KEY_ID': 'blah',
        'AWS_CONFIG_FILE': 'blah',
        'AWS_CREDENTIAL_FILE': 'blah',
        'AWS_DEFAULT_PROFILE': 'blah',
        'AWS_DEFAULT_REGION': 'blah',
        'AWS_EXECUTION_ENV': 'blah',
        'AWS_PROFILE': 'blah',
        'AWS_REGION': 'blah',
        'AWS_SECRET_ACCESS_KEY': 'blah',
        'AWS_SECURITY_TOKEN': 'blah',
        'AWS_SESSION_TOKEN': 'blah',
        'BOTO_CONFIG': 'blah',
    })
    def test_botocore_config_check(self, exists_mock):
        from arkindex.project.checks import botocore_config_check

        expected_warnings = [
            Warning(
                'The AWS_ACCESS_KEY_ID environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_CONFIG_FILE environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_CREDENTIAL_FILE environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_DEFAULT_PROFILE environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_DEFAULT_REGION environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_EXECUTION_ENV environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_PROFILE environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_REGION environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_SECRET_ACCESS_KEY environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_SECURITY_TOKEN environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The AWS_SESSION_TOKEN environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The BOTO_CONFIG environment variable is set, and not all S3 settings are set.\n'
                'This variable might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The ~/.aws/config file exists, and not all S3 settings are set.\n'
                'The settings in this file might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
            Warning(
                'The ~/.aws/credentials file exists, and not all S3 settings are set.\n'
                'The settings in this file might override the Arkindex settings and cause unexpected behavior.',
                id='arkindex.W011',
            ),
        ]

        settings.INGEST_S3_ACCESS_KEY = None
        settings.INGEST_S3_SECRET_KEY = None
        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = None
        settings.AWS_ACCESS_KEY = None
        settings.AWS_SECRET_KEY = None
        settings.AWS_ENDPOINT = None
        settings.AWS_REGION = None

        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_ACCESS_KEY = 'something'
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_SECRET_KEY = 'something'
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_ACCESS_KEY = 'something'
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_SECRET_KEY = 'something'
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.AWS_ENDPOINT = 'something'
        self.assertCountEqual(botocore_config_check(), expected_warnings)

        settings.INGEST_S3_ENDPOINT = 'something'
        # We have set all the required settings here, so this check is skipped.
        self.assertCountEqual(botocore_config_check(), [])

        # The check requires either the endpoint or the region
        settings.AWS_ENDPOINT = None
        settings.AWS_REGION = 'something'
        self.assertCountEqual(botocore_config_check(), [])

        settings.INGEST_S3_ENDPOINT = None
        settings.INGEST_S3_REGION = 'something'
        self.assertCountEqual(botocore_config_check(), [])

    @override_settings()
    def test_transkribus_checks(self):
        from arkindex.process.models import Repository, Revision, Worker, WorkerType, WorkerVersion
        from arkindex.project.checks import transkribus_check

        repo = Repository.objects.create(url='http://fakery')
        tr_type = WorkerType.objects.create(slug='import', display_name='Import')
        transkribus_worker = Worker.objects.create(
            repository=repo,
            name='Transkribus Import',
            slug='transkribus_import',
            type=tr_type
        )
        transkribus_rev = Revision.objects.create(
            hash='1234',
            message='commit commit',
            author='Wolpertinger',
            repo=repo,
        )
        transkribus_worker_version = WorkerVersion.objects.create(
            worker=transkribus_worker,
            revision=transkribus_rev,
            configuration={}
        )

        with self.settings(
            ARKINDEX_FEATURES={
                'transkribus': True
            },
            TRANSKRIBUS_EMAIL='mail@mail.com',
            TRANSKRIBUS_PASSWORD='passpass',
            TRANSKRIBUS_WORKER_VERSION=None
        ):
            self.assertCountEqual(transkribus_check(), [
                Error(
                    'No Transkribus worker version is set; all Transkribus-related features will fail.',
                    hint='settings.TRANSKRIBUS_WORKER_VERSION = None',
                    id='arkindex.E012',
                )
            ])

        with self.settings(
            ARKINDEX_FEATURES={
                'transkribus': True
            },
            TRANSKRIBUS_EMAIL='mail@mail.com',
            TRANSKRIBUS_PASSWORD='passpass',
            TRANSKRIBUS_WORKER_VERSION='bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'
        ):
            self.assertCountEqual(transkribus_check(), [
                Error(
                    'Transkribus worker version does not exist.',
                    hint="settings.TRANSKRIBUS_WORKER_VERSION = 'bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'",
                    id='arkindex.E012',
                )
            ])

        with self.settings(
            ARKINDEX_FEATURES={
                'transkribus': True
            },
            TRANSKRIBUS_EMAIL='mail@mail.com',
            TRANSKRIBUS_PASSWORD='passpass',
            TRANSKRIBUS_WORKER_VERSION=transkribus_worker_version.id
        ):
            self.assertListEqual(transkribus_check(), [])

        with self.settings(
            ARKINDEX_FEATURES={
                'transkribus': False
            },
            TRANSKRIBUS_EMAIL='mail@mail.com',
            TRANSKRIBUS_PASSWORD='passpass',
            TRANSKRIBUS_WORKER_VERSION=None
        ):
            self.assertListEqual(transkribus_check(), [])
